@@ -4,8 +4,7 @@
 
 #if defined(EXTUNIX_HAVE_ATFILE)
 
-#include <string.h>
-#include <errno.h>
+#include "common.h"
 
 /* otherlibs/unix/cst2constr.h */
 extern value cst_to_constr(int n, int * tbl, int size, int deflt);
@@ -126,33 +125,13 @@ CAMLprim value caml_extunix_symlinkat(value v_path, value v_newdirfd, value v_ne
   CAMLreturn(Val_unit);
 }
 
-/* otherlibs/unix/open.c */
-
-#ifndef O_NONBLOCK
-#define O_NONBLOCK O_NDELAY
-#endif
-#ifndef O_DSYNC
-#define O_DSYNC 0
-#endif
-#ifndef O_SYNC
-#define O_SYNC 0
-#endif
-#ifndef O_RSYNC
-#define O_RSYNC 0
-#endif
-
-static int open_flag_table[] = {
-  O_RDONLY, O_WRONLY, O_RDWR, O_NONBLOCK, O_APPEND, O_CREAT, O_TRUNC, O_EXCL,
-  O_NOCTTY, O_DSYNC, O_SYNC, O_RSYNC
-};
-
 CAMLprim value caml_extunix_openat(value v_dirfd, value path, value flags, value perm)
 {
   CAMLparam4(v_dirfd, path, flags, perm);
   int ret, cv_flags;
   char * p;
 
-  cv_flags = caml_convert_flag_list(flags, open_flag_table);
+  cv_flags = extunix_open_flags(flags);
   p = caml_stat_alloc(caml_string_length(path) + 1);
   strcpy(p, String_val(path));
   /* open on a named FIFO can block (PR#1533) */
