@@ -9,6 +9,7 @@ type arg =
   | S of string (* check symbol available *)
   | V of string (* check value available (enum) *)
   | D of string (* check symbol defined *)
+  | F of string * string (* check structure type available and specified field present in it *)
 
 type test =
   | L of arg list
@@ -18,15 +19,8 @@ type t = YES of (string * arg list) | NO of string
 
 let ocamlc = ref "ocamlc"
 let ext_obj = ref ".o"
-let verbose = ref false
-
-let () =
-  let args = [
-    "-ocamlc", Arg.Set_string ocamlc, "<path> ocamlc";
-    "-ext_obj", Arg.Set_string ext_obj, "<ext> C object files extension";
-    "-v", Arg.Set verbose, " Show code for failed tests";
-  ] in
-  Arg.parse args (failwith) ("Options are:")
+let verbose = ref 1
+let disabled = ref []
 
 let print_define b s = bprintf b "#define %s\n" s
 let print_include b s = bprintf b "#include <%s>\n" s
@@ -35,13 +29,14 @@ let get_defines = filter_map (function DEFINE s -> Some s | _ -> None)
 let get_includes = filter_map (function I s -> Some s | _ -> None)
 
 let config_defines = [
-  "_POSIX_C_SOURCE 200112L";
-  "_XOPEN_SOURCE 600";
+  "_POSIX_C_SOURCE 200809L";
+  "_XOPEN_SOURCE 700";
   "_BSD_SOURCE";
   "_LARGEFILE64_SOURCE";
   "WIN32_LEAN_AND_MEAN";
   "_WIN32_WINNT 0x500";
   "CAML_NAME_SPACE";
+  "_GNU_SOURCE"
   ]
 
 let config_includes = [
@@ -51,6 +46,9 @@ let config_includes = [
   "caml/signals.h";
   "caml/alloc.h";
   "caml/custom.h";
+  "caml/bigarray.h";
+  "string.h";
+  "errno.h";
   ]
 
 let build_code args =
@@ -69,17 +67,20 @@ let build_code args =
     | D s -> pr "#ifndef %s" s; pr "#error %s not defined" s; pr "#endif"
     | S s -> pr "size_t var_%d = (size_t)&%s;" (fresh ()) s
     | V s -> pr "int var_%d = (0 == %s);" (fresh ()) s
+    | F (s,f) -> pr "size_t var_%d = (size_t)&((struct %s*)0)->%s" (fresh ()) s f
     end args;
   pr "int main() { return 0; }";
   Buffer.contents b
 
+let dev_null = match Sys.os_type with "Win32" -> "NUL" | _ -> "/dev/null"
+
 let execute code =
   let (tmp,ch) = Filename.open_temp_file "discover" ".c" in
   output_string ch code;
   flush ch;
-  let cmd = sprintf "%s -c %s" !ocamlc (Filename.quote tmp) in
-  let ret = Sys.command cmd in
   close_out ch;
+  let cmd = sprintf "%s -c %s%s" !ocamlc (Filename.quote tmp) (if !verbose >= 1 then "" else " 2> " ^ dev_null) in
+  let ret = Sys.command cmd in
   Sys.remove tmp;
   (* assumption: C compiler puts object file in current directory *)
   let base = Filename.chop_extension (Filename.basename tmp) in
@@ -92,11 +93,14 @@ let discover (name,test) =
     let code = build_code args in
     match execute code, other with
     | false, [] -> 
-        if !verbose then prerr_endline code;
+        if !verbose >= 2 then prerr_endline code;
         print_endline "failed"; NO name
     | false, (x::xs) -> loop x xs
     | true, _ -> print_endline "ok"; YES (name,args)
   in
+  match List.mem name !disabled with
+  | true -> print_endline "disabled"; NO name
+  | false ->
   match test with
   | L l -> loop l []
   | ANY (x::xs) -> loop x xs
@@ -106,6 +110,8 @@ let show_c file result =
   let b = Buffer.create 10 in
   let pr fmt = ksprintf (fun s -> Buffer.add_string b (s^"\n")) fmt in
   pr "";
+  pr "#define UNUSED(x) (void)(x)";
+  pr "";
   List.iter (print_define b) config_defines;
   List.iter begin function
     | NO _ -> ();
@@ -154,8 +160,7 @@ let main config =
   show_c "src/config.h" result;
   show_ml "src/config.ml" result
 
-let () = 
-  main 
+let features =
   [
     "EVENTFD", L[
       I "sys/eventfd.h";
@@ -225,7 +230,7 @@ let () =
       [D "WIN32"; S "FlushFileBuffers"; ];
     ];
     "REALPATH", L[ I "limits.h"; I "stdlib.h"; S "realpath"; ];
-    "SIGNALFD", L[ I "sys/signalfd.h"; S "signalfd"; I "signal.h"; S "sigemptyset"; S "sigaddset"; I "errno.h"; D "EINVAL"; ];
+    "SIGNALFD", L[ I "sys/signalfd.h"; S "signalfd"; I "signal.h"; S "sigemptyset"; S "sigaddset"; ];
     "PTRACE", L[ I "sys/ptrace.h"; S "ptrace"; V "PTRACE_TRACEME"; V "PTRACE_ATTACH"; ];
     "RESOURCE", L[
       I "sys/time.h"; I "sys/resource.h";
@@ -244,5 +249,43 @@ let () =
     "SETENV", L[ I"stdlib.h"; S"setenv"; S"unsetenv"; ];
     "CLEARENV", L[ I"stdlib.h"; S"clearenv"; ];
     "MKDTEMP", L[ I"stdlib.h"; S"mkdtemp"; ];
+    "TIMEGM", L[ I"time.h"; S"timegm"; ];
+    "MALLOC_INFO", L[ I"malloc.h"; S"malloc_info"; ];
+    "MALLOC_STATS", L[ I"malloc.h"; S"malloc_stats"; ];
+    "MEMALIGN", L[ I "stdlib.h"; S"posix_memalign"; ];
+    "ENDIAN", L[ I "endian.h";
+		 D"htobe16"; D"htole16"; D"be16toh"; D"le16toh";
+		 D"htobe32"; D"htole32"; D"be32toh"; D"le32toh";
+		 D"htobe64"; D"htole64"; D"be64toh"; D"le64toh"; ];
+    "READ_CREDENTIALS", L[ I"sys/types.h"; I"sys/socket.h"; D"SO_PEERCRED"; ];
+    "FEXECVE", L[ I "unistd.h"; S"fexecve"; ];
+    "SENDMSG", ANY[
+      [ I"sys/types.h"; I"sys/socket.h"; S"sendmsg"; S"recvmsg"; D"CMSG_SPACE"; ];
+      [ I"sys/types.h"; I"sys/socket.h"; S"sendmsg"; S"recvmsg"; F("msghdr","msg_accrights"); ];
+    ];
+    "PREAD", L[ I "unistd.h"; S"pread"; ];
+    "PWRITE", L[ I "unistd.h"; S"pwrite"; ];
+    "READ", L[ I "unistd.h"; S"read"; ];
+    "WRITE", L[ I "unistd.h"; S"write"; ];
+    "MKSTEMPS", L[ I "stdlib.h"; S"mkstemps"; ];
+    "MKOSTEMPS", L[ I "stdlib.h"; S"mkostemps"; ];
+    "SETRESUID", L[ I"sys/types.h"; I"unistd.h"; S"setresuid"; S"setresgid" ];
   ]
 
+let () = 
+  let args0 = [
+    "-ocamlc", Arg.Set_string ocamlc, "<path> ocamlc";
+    "-ext_obj", Arg.Set_string ext_obj, "<ext> C object files extension";
+    "-v", Arg.Unit (fun () -> verbose := 2), " Show code for failed tests";
+    "-q", Arg.Unit (fun () -> verbose := 0), " Do not show stderr from children";
+  ] in
+  let args1 = List.map (fun (name,_) ->
+    assert (not (String.contains name ' '));
+    "--disable-" ^ String.lowercase name,
+    Arg.Unit (fun () -> disabled := name :: !disabled),
+    " disable " ^ name) features 
+  in
+  let args = Arg.align (args0 @ args1) in
+  Arg.parse args failwith ("Options are:");
+  main features
+
